######################################################################################################
##
##  Reference implementation of the KSN format for GNU R
##
##  Copyright (c) 2009-2011 Stanisław A. Raczyński,
##                          Sagayama Laboratory,
##                          The University of Tokyo
##
######################################################################################################

ksn2Harte = function(keylabel, chordlabel)
{
  key   = interpretKSKey(keylabel, FALSE)
  chord = interpretKSChord(chordlabel, tonic=key["Tonic"], mode=key["Mode"])
  key   = interpretKSKey(keylabel, TRUE)

  tonic = .keySignatures[key["Mode"] + 1, key["Signature"] + 8]
  if(key["Mode"] == 0)
    degrees = .majorDegrees
  else
    degrees = .minorNaturalDegrees

  if(chord["Degree"] > 0)
  {
    addDegree = function(noteDegree)
    {
      if(is.na(chord[.degreeNames[noteDegree]]))
        return(NA)
      scaleDegree = chord["Degree"] + noteDegree - 1
      scaleDegree = (scaleDegree - 1) %% 7 + 1
      if(key["Mode"] == 1 && (scaleDegree %in% c(6,7)))   # for minor scales the sixth or the seventh can be raised
      {
        noteCandidates = c(.minorNaturalDegrees[scaleDegree], .minorMelodicDegrees[scaleDegree]) + chord[.degreeNames[noteDegree]]
        return(noteCandidates[which(noteCandidates %in% .standardIntervals)]  - rootPos)
      }
      else  
        return(degrees[scaleDegree] + chord[.degreeNames[noteDegree]] - rootPos)
    }

    root = tonic + degrees[chord["Degree"]]
    if(root > 11)
      root = root - 12
    if(key["Signature"] >= 0)
      harteLabel = .pitchNamesSharp.Harte[root + 1]
    else
      harteLabel = .pitchNamesFlat.Harte[root + 1]
    missingRoot = is.na(chord["Root"])
    harteLabel = paste(harteLabel, ":", sep="")

    intervals = rep(NA, 15)
    rootPos = degrees[(chord["Degree"] - 1) %% 7 + 1]
    for(degree in 2:15)
      intervals[degree] = addDegree(degree)
    intervals[!is.na(intervals) & intervals < 0] = intervals[!is.na(intervals) & intervals < 0] + 12

    # Make sure the intervals are in ascending pitch order
    maxInterval = 0
    for(intervalNr in which(!is.na(intervals)))
    {
      if(intervals[intervalNr] < maxInterval)
        intervals[intervalNr] = intervals[intervalNr] + 12  # shift them up by an octave
      maxInterval = intervals[intervalNr]
    }
    intervals[1] = chord["Root"]

    # Correct the third according to the chord type
    if(chord["Type"] == 0)  # major chord
    {
      if(!is.na(chord[.degreeNames["Third"]]) && intervals[3] != 4)
        warning("Out of key third.")
      intervals[3] = 4
    }
    else   # minor chord
    {
      if(!is.na(chord[.degreeNames["Third"]]) && intervals[3] != 3)
        warning("Out of key third.")
      intervals[3] = 3
    }

    # Note it down if the third is missing
    if(is.na(chord["Third"]))
      missingThird = paste("*", .modifiers.Harte[(intervals[3] - .standardIntervals[3]) + 3], 3, sep="")
    else
      missingThird = ""
      
    # See if the chord is of any of the common types
    chordType = ""
    if(is.na(intervals[1]) || (intervals[1] == 0))
    {
      if(!is.na(intervals[3]) && !is.na(intervals[5]) && !is.na(intervals[7]) && !is.na(intervals[9]))   # Ninth chords
      {
        if(intervals[3] == 4 && intervals[5] == 7 && intervals[7] == 10 && intervals[9] == 14)
          chordType = "9"
        else if(intervals[3] == 4 && intervals[5] == 7 && intervals[7] == 11 && intervals[9] == 14)
         chordType = "maj9"
        else if(intervals[3] == 3 && intervals[5] == 7 && intervals[7] == 10 && intervals[9] == 14)
          chordType = "min9"
        if(chordType != "")
          intervals[c(1,3,5,7,9)] = NA  # remove the notes included in the chordType
      }
      else if(!is.na(intervals[3]) && !is.na(intervals[5]) && !is.na(intervals[7])) # Seventh chords
      {
        if(intervals[3] == 4 && intervals[5] == 7 && intervals[7] == 11)
          chordType = "maj7"
        else if(intervals[3] == 3 && intervals[5] == 7 && intervals[7] == 10)
         chordType = "min7"
        else if(intervals[3] == 4 && intervals[5] == 7 && intervals[7] == 10)
          chordType = "7"
        else if(intervals[3] == 3 && intervals[5] == 6 && intervals[7] == 9)
          chordType = "dim7"
        else if(intervals[3] == 3 && intervals[5] == 6 && intervals[7] == 10)
          chordType = "hdim7"
        else if(intervals[3] == 3 && intervals[5] == 7 && intervals[7] == 11)
          chordType = "minmaj7"
        if(chordType != "")
          intervals[c(1,3,5,7)] = NA  # remove the notes included in the chordType
      }
      else if(!is.na(intervals[3]) && !is.na(intervals[5]) && !is.na(intervals[6])) # Sixths chords
      {
        if(intervals[3] == 4 && intervals[5] == 7 && intervals[6] == 9)
          chordType = "maj6"
        else if(intervals[3] == 3 && intervals[5] == 7 && intervals[6] == 9)
          chordType = "min6"
        if(chordType != "")
          intervals[c(1,3,5,6)] = NA  # remove the notes included in the chordType
      }
      else if(!is.na(intervals[4]) && !is.na(intervals[5])) # Suspended 4th chords
      {
        if(intervals[4] == 5 && intervals[5] == 7)
          chordType = "sus4"
        if(chordType != "")
          intervals[c(1,4,5)] = NA  # remove the notes included in the chordType
      }
      else if(!is.na(intervals[2]) && !is.na(intervals[5])) # Suspended 4th chords
      {
        if(intervals[2] == 2 && intervals[5] == 7)
          chordType = "sus2"
        if(chordType != "")
          intervals[c(1,2,5)] = NA  # remove the notes included in the chordType
      }
      else if(!is.na(intervals[5])) # Triads
      {
        if(intervals[3] == 4 && intervals[5] == 7)
          chordType = "maj"
        else if(intervals[3] == 3 && intervals[5] == 7)
          chordType = "min"
        else if(intervals[3] == 3 && intervals[5] == 6)
          chordType = "dim"
        else if(intervals[3] == 4 && intervals[5] == 8)
          chordType = "aug"
        if(chordType != "")
          intervals[c(1,3,5)] = NA  # remove the notes included in the chordType
      }
    }
    
    if(chordType != "")
      harteLabel = paste(harteLabel, chordType, sep="")

    # Explicit degree list, added in parentheses
    if(missingThird != "")
      intervals[3] = NA
    degreeList = c()
    for(degree in which(!is.na(intervals)))
      degreeList = c(degreeList, paste(.modifiers.Harte[(intervals[degree] - .standardIntervals[degree]) + 3], degree, sep=""))
    if(missingThird != "" && (chordType != "" || length(degreeList) == 0))
      degreeList = c(missingThird, degreeList)
    if(missingRoot)
      degreeList = c("*1", degreeList)
    if(length(degreeList) > 0)
      harteLabel = paste(harteLabel, "(", paste(degreeList, collapse=","), ")", sep="")
  }
  else
    harteLabel = "N"  # non-degree chords cannot be used in Harte notation

  return(harteLabel)
}

ksn2lab = function(ksnfilename, barfilename, labfilename, songlength)
{
  barPositions = read.table(barfilename, header=FALSE, sep=",")
  KSData = readKSFile(ksnfilename, ignoreloops=FALSE)

  measures = cumsum(c(1, (KSData[,1] / KSData[,5]) / KSData[,4]))

  measure2time = function(measureNr)
  {
    wholeBars    = floor(measureNr)
    fractionBars = measureNr - wholeBars
    if(wholeBars > nrow(barPositions))
      timePos = songlength
    else if(wholeBars == nrow(barPositions))
      timePos = barPositions[wholeBars, 1] + (songlength - barPositions[wholeBars, 1]) * fractionBars
    else
      timePos = barPositions[wholeBars, 1] + (barPositions[wholeBars+1, 1] - barPositions[wholeBars, 1]) * fractionBars
    return(timePos)
  }

  timings = matrix(NA, nrow(KSData), 2)
  labels  = rep(NA, nrow(KSData))
  for(chordLabelNr in 1:nrow(KSData))
  {
    timings[chordLabelNr,] = c(measure2time(measures[chordLabelNr]), measure2time(measures[chordLabelNr+1]))
    labels[chordLabelNr]   = paste(ksn2Harte(KSData[chordLabelNr, "Key"], KSData[chordLabelNr, "Chord"]), sep="")
  }

  harteLabels = data.frame(timings, labels)
  write.table(harteLabels, labfilename, col.names=FALSE, row.names=FALSE, sep=" ", quote=FALSE)
}

