######################################################################################################
##
##  Reference implementation of the KSN format for GNU R
##
##  Copyright (c) 2009-2011 Stanisław A. Raczyński,
##                          Sagayama Laboratory,
##                          The University of Tokyo
##
######################################################################################################

######################################################################################################
## ksn2midi() reads a KSN file and converts it into a MIDI format
ksn2midi = function(ksnfilename, midifilename, BPM=120, basicNote=60, enharmonic=TRUE, ignoreloops=FALSE)
                                                        # 48 = C3, 60 = C4
{
  ticksPerBeat = 384

  chunkHeader = function(ID, len)  # creates a chuck header data
  {
    byteData = rep(0, 4)
    byteData[1] = floor(len/0x01000000)
    byteData[2] = floor((len - byteData[1] * 0x01000000) / 0x00010000)
    byteData[3] = floor((len - byteData[1] * 0x01000000 - byteData[2] * 0x00010000) / 0x00000100)
    byteData[4] = (len - byteData[1] * 0x01000000 - byteData[2] * 0x00010000 - byteData[3] * 0x00000100)
    return(c(charToRaw(ID), as.vector(byteData, mode="raw")))
  }

  generateVariableQuantity = function(value)
  {
    variableQuantity <- vector(0, mode="raw")
    data = rep(0, 4)
    data[1] = floor(value/(128*128*128))  # MSB
    data[2] = floor((value - data[1] * (128*128*128)) / (128*128))
    data[3] = floor((value - data[1] * (128*128) - data[2] * (128*128)) / (128))
    data[4] = (value - data[1] * (128*128*128) - data[2] * (128*128) - data[3] * (128)) # LSB
    someAlreadyWritten = FALSE
    for(byteNr in 1:4)
    {
      if((data[byteNr] > 0) || someAlreadyWritten)
      {
        variableQuantity = c(variableQuantity, as.raw(data[byteNr]))
        someAlreadyWritten = TRUE
      }
    }
    if(!someAlreadyWritten)
      variableQuantity = c(variableQuantity, as.raw(0))
    if(length(variableQuantity) > 1)
      for(byteNr in 1:(length(variableQuantity)-1))
        variableQuantity[byteNr] = as.raw(as.numeric(variableQuantity[byteNr]) + 128)  # set the MSB
    return(variableQuantity)
  }

  addNoteOnEvent = function(trackData, note, deltaTime=0)
  {
    trackData = c(trackData, generateVariableQuantity(deltaTime))
    return(c(trackData, as.vector(c(0x90, note, 0x7F), mode="raw"))) # note ON, channel 0
  }

  addNoteOffEvent = function(trackData, note, deltaTime=0)
  {
    trackData = c(trackData, generateVariableQuantity(deltaTime))
    return(c(trackData, as.vector(c(0x80, note, 0x7F), mode="raw"))) # note OFF, channel 0
  }

  addTempoEvent = function(trackData, BPM=120, deltaTime=0)  # tempo is beats per minute
  {
    msPerQuarterNote = 60000000 / BPM
    byteData = rep(0, 3)
    byteData[1] = floor(msPerQuarterNote / 0x00010000)
    byteData[2] = floor((msPerQuarterNote - byteData[1] * 0x00010000) / 0x00000100)
    byteData[3] = msPerQuarterNote - byteData[1] * 0x00010000 - byteData[2] * 0x00000100
    trackData = c(trackData, generateVariableQuantity(deltaTime))
    return(c(trackData, as.vector(c(0xFF, 0x51, 3, byteData), mode="raw")))
  }

  addEndOfTrackEvent = function(trackData, deltaTime=ticksPerBeat)
  {
    trackData = c(trackData, generateVariableQuantity(deltaTime)) # by default, track will end 1 beat after the last event
    return(c(trackData, as.vector(c(0xFF, 0x2F, 0x00), mode="raw")))  # End-of-track marker
  }

  addTimeSignatureEvent = function(trackData, beatsPerMeasure, ticksPerBeat, deltaTime=0)
  {
    beatValue = round(log2(96 / ticksPerBeat))
    trackData = c(trackData, generateVariableQuantity(deltaTime))
    return(c(trackData, as.vector(c(0xFF, 0x58, 4, beatsPerMeasure, beatValue, 24, 8), mode="raw")))
  }

  addKeySignatureEvent = function(trackData, keySignature, keyMode, deltaTime=0)
  {
      # these are two's complement representations of numbers -7...7
      # we have to explicitly list them because as far as I know, R does not support
      # raw _signed_ character vectors
    binarySignatures = c(0xF9, 0xFA, 0xFB, 0xFC, 0xFD, 0xFE, 0xFF, 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07)
    trackData = c(trackData, generateVariableQuantity(deltaTime))
    return(c(trackData, as.vector(c(0xFF, 0x59, 2, binarySignatures[keySignature+8], keyMode), mode="raw")))
  }

  addLyricsEvent = function(trackData, lyrics, deltaTime=0)
  {
    trackData = c(trackData, generateVariableQuantity(deltaTime))
    return(c(trackData, c(as.vector(c(0xFF, 0x05, nchar(lyrics)), mode="raw"), charToRaw(lyrics))))
  }

  addTrackNameEvent = function(trackData, trackName, deltaTime=0)
  {
    trackData = c(trackData, generateVariableQuantity(deltaTime))
    return(c(trackData, c(as.vector(c(0xFF, 0x03, nchar(trackName)), mode="raw"), charToRaw(trackName))))
  }

  cat("Reading harmony...\n")
  KSData = readKSFile(ksnfilename, ignoreloops=ignoreloops)
  harmonyData = matrix(NA, nrow(KSData), 3+2+18)
  colnames(harmonyData) = c("Measures", "Beats", "Ticks", names(interpretKSKey(KSData[1, 2], enharmonic)), names(interpretKSChord(KSData[1, 3])))
  lengths = KSData[,1]
  lengths = cbind(lengths / KSData[,5], lengths)
  lengths = cbind(lengths[,1] / KSData[,4], lengths)
  harmonyData[, 1:3] = lengths
  for(ii in 1:nrow(harmonyData))
  {
    key = interpretKSKey(KSData[ii, 2], FALSE)
    harmonyData[ii, 4:ncol(harmonyData)] = c(interpretKSKey(KSData[ii, 2], enharmonic), interpretKSChord(KSData[ii, 3], tonic=key["Tonic"], mode=key["Mode"]))
  }

  cat("Generating MIDI data... ")
  timeSignature = c(harmonyData[1, "Beats"] / harmonyData[1, "Measures"], harmonyData[1, "Ticks"] / harmonyData[1, "Beats"])
  trackData = addTempoEvent(vector(length=0, mode="raw"), BPM)
  trackData = addTimeSignatureEvent(trackData, timeSignature[1], timeSignature[2])
  trackData = addKeySignatureEvent(trackData, harmonyData[1, "Signature"], harmonyData[1, "Mode"])
  trackData = addTrackNameEvent(trackData, "Harmony")
  previousEventTime = 0
  deltaTime = 0
  cumTime = c(0, cumsum(harmonyData[, "Beats"]))  # absolute time of harmony changes (in beats)
  for(harmonyNr in 1:nrow(harmonyData))
  {
    # Chord labels
    trackData = addLyricsEvent(trackData, paste(KSData[harmonyNr, "Key"], ":", KSData[harmonyNr, "Chord"], sep=""), deltaTime)

    # Time signature changes
    currentTimeSignature = c(harmonyData[harmonyNr, "Beats"] / harmonyData[harmonyNr, "Measures"], harmonyData[harmonyNr, "Ticks"] / harmonyData[harmonyNr, "Beats"])
    if(any(currentTimeSignature != timeSignature))
    {
      timeSignature = currentTimeSignature
      trackData = addTimeSignatureEvent(trackData, timeSignature[1], timeSignature[2])
    }

    # Chord notes
    notes = harmony2notes(harmonyData[harmonyNr,], basicNote=basicNote)
    for(note in notes)
      trackData = addNoteOnEvent(trackData, note, deltaTime)
    deltaTime = (cumTime[harmonyNr+1] - previousEventTime) * ticksPerBeat
    previousEventTime = cumTime[harmonyNr+1]
    for(note in notes)
    {
      trackData = addNoteOffEvent(trackData, note, deltaTime)
      deltaTime = 0
    }
  }
  trackData = addEndOfTrackEvent(trackData)
  cat(length(trackData), " bytes \n", sep="")

  cat("Saving MIDI data...\n")
  MIDIfile = file(midifilename, "wb")
  writeBin(chunkHeader("MThd", 6), MIDIfile)
  writeBin(as.vector(c(0,0, 0,1, 0x01,0x80), mode="raw"), MIDIfile)
    # 0,0    = format type 0 (single track MIDI)
    # 0,1    = number of tracks (1)
    # 0x0180 = 384 ticks per beat
  writeBin(chunkHeader("MTrk", length(trackData)), MIDIfile)
  writeBin(trackData, MIDIfile)
  close(MIDIfile)
}

